"""Inquisitor sessions."""
import json
from datetime import timedelta
from typing import AsyncGenerator, Optional

from fastapi import Depends, Security
from fred_types import BaseModel
from oic import rndstr
from pydantic import Field
from redis.asyncio import Redis
from verifier import ContactData, VerificationId, VerificationResult

from .constants import Scope
from .models import Error
from .settings import Settings, get_cache, get_settings
from .token import TokenData, auth_token_data


class OIDCSession(BaseModel):
    """OIDC session."""

    redirect_url: str = "/"
    state: str = Field(default_factory=rndstr)
    nonce: str = Field(default_factory=rndstr)
    access_token: str | None = None
    mojeid_nia: bool = False
    mojeid_validation: bool = False
    mojeid_username: str = ""
    contact_data: ContactData | None = None
    verification_result: VerificationResult | None = None
    error: Error | None = None

    @property
    def is_legal_person(self) -> bool:
        """Determine whether contact is a legal person."""
        assert self.contact_data is not None  # nosec: assert_used
        if self.contact_data.organization:
            assert (
                self.mojeid_nia is False
            )  # legal person cannot be attached to NIA  # nosec: assert_used
        return bool(self.contact_data.organization)

    @property
    def is_trustworthy(self) -> bool:
        """Determine whether contact data are trustworthy.

        Legal person data are trustworthy if and only if it's validated.
        Physical person data are trustworthy if and only if it's attached to NIA.
        """
        if self.is_legal_person:
            return self.mojeid_validation
        else:
            return self.mojeid_nia


class Session(BaseModel):
    """Session model."""

    verification_id: VerificationId
    oidc: OIDCSession = Field(default_factory=OIDCSession)
    representative_token: Optional[str] = None


def _cache_key(verification_id: VerificationId, settings: Settings) -> str:
    """Get key to session data."""
    return f"{settings.redis_cache_key_prefix}session:{verification_id}"


async def create_session(session: Session, cache: Redis, settings: Settings) -> None:
    """Create session object."""
    cache_key = _cache_key(session.verification_id, settings)
    await cache.set(
        cache_key,
        session.json(),
        ex=timedelta(seconds=settings.access_token_expires),
    )


async def get_session(
    cache: Redis = Depends(get_cache),
    settings: Settings = Depends(get_settings),
    token: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> AsyncGenerator[Session, None]:
    """Get session object.

    Session content is always saved before sending response.
    """
    # This code is covered by tests, but we need to ignore coverage
    # because of Python 3.11 coverage bug.
    # See https://github.com/nedbat/coveragepy/issues/1595
    cache_key = _cache_key(token.verification_id, settings)
    session_data = await cache.get(cache_key)
    if session_data is None:  # pragma: no cover
        session = Session(verification_id=token.verification_id)
    else:  # pragma: no cover
        session = Session(**json.loads(session_data))

    try:  # pragma: no cover
        yield session
    finally:
        await cache.set(
            cache_key,
            session.json(),
            ex=timedelta(seconds=settings.access_token_expires),
        )
