from unittest import TestCase
from unittest.mock import call

from fastapi.testclient import TestClient
from fred_types import ContactId
from freezegun import freeze_time
from jose import jwe
from verifier import VerificationId
from verifier.exceptions import VerificationDoesNotExist

from inquisitor.constants import OAUTH_COOKIE, LogEntryType, LoggerResult, LogReferenceType, Scope
from inquisitor.main import app
from inquisitor.settings import get_settings
from inquisitor.token import TokenData, create_access_token

from .utils import LoggerMixin, VerificationClientMixin, make_verification

client = TestClient(app)


class TestGetAccessToken:
    @freeze_time("2022-01-01T00:00:00Z")
    class Base(VerificationClientMixin, LoggerMixin, TestCase):
        url: str

        def assertTokensEqual(self, t1, t2):
            self.assertEqual(
                jwe.decrypt(t1, get_settings().secret_key_sha256),
                jwe.decrypt(t2, get_settings().secret_key_sha256),
            )

        def _test_login(self, credentials, granted_scopes):
            self.verification_client.get.return_value = make_verification(
                verification_id="v:1", contact_id="c:1"
            )

            response = client.post(self.url, data=credentials)

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), {"scopes": granted_scopes})
            self.assertTokensEqual(
                response.cookies[OAUTH_COOKIE].removeprefix('"Bearer ').removesuffix('"'),
                create_access_token(
                    TokenData(
                        verification_id=VerificationId("v:1"),
                        contact_id=ContactId("c:1"),
                        scopes=granted_scopes,
                    ),
                    settings=get_settings(),
                ),
            )

        def test_login_with_password(self):
            self._test_login({"username": "v:1", "password": "starbug"}, [Scope.READ, Scope.WRITE])
            self.assertEqual(
                self.verification_client.mock_calls,
                [call.get(VerificationId("v:1"), access_token="starbug")],
            )
            self.assertLog(
                LogEntryType.LOGIN,
                LoggerResult.SUCCESS,
                input_properties={"username": "v:1", "password": str(True), "scopes": []},
                output_properties={"granted_scopes": [Scope.READ, Scope.WRITE]},
                output_references={
                    LogReferenceType.VERIFICATION: "v:1",
                    LogReferenceType.CONTACT: "c:1",
                },
            )

        def test_wrong_credentials(self):
            self.verification_client.get.side_effect = VerificationDoesNotExist

            response = client.post(self.url, data={"username": "v:1", "password": "psirens"})

            self.assertEqual(response.json(), {"detail": "Incorrect username or password"})
            self.assertEqual(response.status_code, 400)
            self.assertLog(
                LogEntryType.LOGIN,
                LoggerResult.FAIL,
                input_properties={"username": "v:1", "password": str(True), "scopes": []},
            )
