from datetime import datetime, timezone
from unittest import TestCase
from unittest.mock import call

from fastapi.testclient import TestClient
from freezegun import freeze_time
from jose import jwe, jwt
from jose.constants import ALGORITHMS
from regal.exceptions import ContactDoesNotExist
from verifier import PaginatedList, VerificationId

from inquisitor.constants import OAUTH_COOKIE, Scope
from inquisitor.main import app
from inquisitor.settings import Settings, get_settings
from inquisitor.token import TokenData, auth_token_data

from .auth import TestGetAccessToken
from .utils import (
    CONTACT_ID,
    VERIFICATION_ID,
    ContactClientMixin,
    LoggerMixin,
    VerificationClientMixin,
    make_verification,
)

client = TestClient(app)


class GetAccessTokenTest(TestGetAccessToken.Base):
    url = "/auth/token/"

    def test_login_without_password(self):
        response = client.post(self.url, data={"username": "v:1"})

        self.assertEqual(response.json(), {"detail": "Password is required"})
        self.assertEqual(response.status_code, 400)
        self.assertNoLogs()


@freeze_time("2022-01-01T00:00:00Z")
class GetRepresentativeAccessTokenTest(
    ContactClientMixin, VerificationClientMixin, LoggerMixin, TestCase
):
    url = "/auth/representative/token/"

    def test_login_without_password(self):
        response = client.post(self.url, data={"username": "RIMMER"})

        self.assertEqual(response.json(), {"detail": "Password is required"})
        self.assertEqual(response.status_code, 400)

    def test_wrong_contact_handle(self):
        self.contact_client.get_contact_id.side_effect = ContactDoesNotExist
        response = client.post(self.url, data={"username": "HOLLY", "password": "starbug"})

        self.assertEqual(response.json(), {"detail": "Incorrect username or password"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("HOLLY")])
        self.assertEqual(self.verification_client.mock_calls, [])

    def test_no_verifications(self):
        self.contact_client.get_contact_id.return_value = "c:1"
        self.verification_client.list.return_value = PaginatedList(results=[])

        response = client.post(self.url, data={"username": "RIMMER", "password": "starbug"})

        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(response.status_code, 404)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("RIMMER")])
        self.assertEqual(self.verification_client.mock_calls, [call.list(contact_id="c:1")])

    def test_multiple_verifications(self):
        self.contact_client.get_contact_id.return_value = "c:1"
        self.verification_client.list.return_value = PaginatedList(
            results=[
                make_verification(
                    verification_id="v:1",
                    create_datetime=datetime(2022, 1, 1, tzinfo=timezone.utc),
                ),
                make_verification(
                    verification_id="v:2",
                    create_datetime=datetime(2021, 1, 1, tzinfo=timezone.utc),
                    close_datetime=datetime(2021, 2, 1, tzinfo=timezone.utc),
                ),
            ]
        )
        self.verification_client.verify_token.return_value = True
        self.verification_client.get.return_value = make_verification()

        response = client.post(self.url, data={"username": "RIMMER", "password": "starbug"})

        self.assertEqual(response.json(), {"scopes": [Scope.READ]})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("RIMMER")])
        self.assertEqual(
            self.verification_client.mock_calls,
            [
                call.list(contact_id="c:1"),
                call.verify_token(VerificationId("v:1"), "starbug", check=True),
                call.get(VerificationId("v:1"), access_token=""),
            ],
        )

    def test_invalid_verify_token(self):
        self.contact_client.get_contact_id.return_value = "c:1"
        self.verification_client.list.return_value = PaginatedList(results=[make_verification()])
        self.verification_client.verify_token.return_value = False

        response = client.post(self.url, data={"username": "RIMMER", "password": "starbug"})

        self.assertEqual(response.json(), {"detail": "Incorrect username or password"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("RIMMER")])
        self.assertEqual(
            self.verification_client.mock_calls,
            [
                call.list(contact_id="c:1"),
                call.verify_token(VerificationId("v:1"), "starbug", check=True),
            ],
        )


class RefreshAccessTokenTest(TestCase):
    def setUp(self):
        self.settings = Settings(secret_key="Gazpacho!", access_token_expires=24 * 60 * 60)

        app.dependency_overrides[get_settings] = lambda: self.settings
        app.dependency_overrides[auth_token_data] = lambda: TokenData(
            verification_id=VERIFICATION_ID,
            contact_id=CONTACT_ID,
            scopes=[Scope.READ, Scope.WRITE],
        )

    def tearDown(self):
        app.dependency_overrides = {}

    @freeze_time("2022-01-01T01:00:00Z")
    def test_refresh_access_token(self):
        response = client.post("/auth/token/refresh/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"scopes": [Scope.READ, Scope.WRITE]})
        access_token = response.cookies[OAUTH_COOKIE].removeprefix('"Bearer ').removesuffix('"')
        decrypted = jwe.decrypt(access_token, self.settings.secret_key_sha256)
        payload = jwt.decode(decrypted, self.settings.secret_key, algorithms=[ALGORITHMS.HS256])
        self.assertEqual(
            payload,
            {
                "verification_id": str(VERIFICATION_ID),
                "contact_id": str(CONTACT_ID),
                "exp": datetime(2022, 1, 2, 1, tzinfo=timezone.utc).timestamp(),
                "iat": datetime(2022, 1, 1, 1, tzinfo=timezone.utc).timestamp(),
                "nbf": datetime(2022, 1, 1, 0, 55, tzinfo=timezone.utc).timestamp(),
                "scopes": [Scope.READ.value, Scope.WRITE.value],
            },
        )


class LogoutTest(TestCase):
    def test_logout(self):
        client = TestClient(app, cookies={OAUTH_COOKIE: "Gazpacho!", "some": "value"})
        response = client.post("/auth/logout/")

        self.assertEqual(response.status_code, 200)
        # There is some bug in starlette that doesn't allow testing response.cookies directly.
        self.assertIn('access_token=""', response.headers["set-cookie"])
