from datetime import datetime, timedelta
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import call

from fastapi import HTTPException, Request
from fastapi.security import SecurityScopes
from fred_types import ContactId
from freezegun import freeze_time
from jose import jwe, jwt
from jose.constants import ALGORITHMS
from verifier import VerificationId
from verifier.exceptions import VerificationDoesNotExist

from inquisitor.constants import OAUTH_COOKIE, Scope
from inquisitor.settings import Settings, get_settings
from inquisitor.token import (
    OAuth2PasswordBearerWithCookie,
    TokenData,
    auth_token_data,
    create_access_token,
    get_contact_id,
    get_verification,
    get_verification_id,
)

from .utils import VerificationClientMixin, make_verification


class OAuth2PasswordBearerWithCookieTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.oauth2_scheme = OAuth2PasswordBearerWithCookie(tokenUrl="auth/token/", scopes={})

    async def test_oauth2_cookie_present(self):
        request = Request(scope={"type": "http", "headers": {}})
        request.cookies[OAUTH_COOKIE] = "Bearer Gazpacho!"
        self.assertEqual(await self.oauth2_scheme(request), "Gazpacho!")

    async def test_oauth2_cookie_missing(self):
        request = Request(scope={"type": "http", "headers": {}})
        with self.assertRaises(HTTPException) as exc:
            await self.oauth2_scheme(request)
        self.assertEqual(exc.exception.status_code, 401)
        self.assertEqual(exc.exception.detail, "Not authenticated")
        self.assertEqual(exc.exception.headers, {"WWW-Authenticate": "Bearer"})


class CreateAccessTokenTest(TestCase):
    @freeze_time("2022-01-01T00:00:00Z")
    def test_create_access_token(self):
        settings = Settings()
        settings.secret_key = "Gazpacho!"
        settings.access_token_expires = int(timedelta(days=1).total_seconds())

        access_token = create_access_token(
            TokenData(
                verification_id=VerificationId("v:1"),
                contact_id=ContactId("c:1"),
                scopes=[Scope.READ],
            ),
            settings=settings,
        )

        decrypted = jwe.decrypt(access_token, settings.secret_key_sha256)
        self.assertEqual(
            jwt.decode(decrypted, settings.secret_key, algorithms=[ALGORITHMS.HS256]),
            {
                "verification_id": "v:1",
                "contact_id": "c:1",
                "scopes": [Scope.READ],
                "nbf": int((datetime.utcnow() - timedelta(minutes=5)).timestamp()),
                "iat": int(datetime.utcnow().timestamp()),
                "exp": int((datetime.utcnow() + timedelta(days=1)).timestamp()),
            },
        )


class AuthTokenDataTest(IsolatedAsyncioTestCase):
    async def test_auth_token_data(self):
        token_data = TokenData(
            verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"), scopes=[Scope.READ]
        )
        access_token = create_access_token(token_data, get_settings())

        self.assertEqual(
            await auth_token_data(
                security_scopes=SecurityScopes([Scope.READ]),
                token=access_token,
                settings=get_settings(),
            ),
            token_data,
        )

    async def test_insufficient_permissions(self):
        token_data = TokenData(
            verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"), scopes=[Scope.READ]
        )
        access_token = create_access_token(token_data, get_settings())

        with self.assertRaises(HTTPException) as exc:
            await auth_token_data(
                security_scopes=SecurityScopes([Scope.WRITE]),
                token=access_token,
                settings=get_settings(),
            )
        self.assertEqual(exc.exception.status_code, 403)
        self.assertEqual(exc.exception.detail, "Insufficient permissions")
        self.assertEqual(exc.exception.headers, {"WWW-Authenticate": 'Bearer scope="write"'})

    async def test_invalid_token(self):
        with self.assertRaises(HTTPException) as exc:
            await auth_token_data(
                security_scopes=SecurityScopes([Scope.WRITE]),
                token="invalid.token",
                settings=get_settings(),
            )
        self.assertEqual(exc.exception.status_code, 401)
        self.assertEqual(exc.exception.detail, "Could not validate credentials")
        self.assertEqual(exc.exception.headers, {"WWW-Authenticate": "Bearer"})


class GetContactIdTest(VerificationClientMixin, IsolatedAsyncioTestCase):
    async def test_get_contact_id(self):
        self.assertEqual(
            await get_contact_id(
                TokenData(verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"))
            ),
            ContactId("c:1"),
        )


class GetVerificationIdTest(VerificationClientMixin, IsolatedAsyncioTestCase):
    async def test_get_verification_id(self):
        self.assertEqual(
            await get_verification_id(
                TokenData(verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"))
            ),
            VerificationId("v:1"),
        )


class GetVerificationTest(VerificationClientMixin, IsolatedAsyncioTestCase):
    async def test_get_verification(self):
        verification = make_verification()
        self.verification_client.get.return_value = verification
        self.assertEqual(
            await get_verification(VerificationId("v:1"), self.verification_client),
            verification,
        )
        self.assertEqual(self.verification_client.mock_calls, [call.get(VerificationId("v:1"))])

    async def test_get_verification_not_found(self):
        self.verification_client.get.side_effect = VerificationDoesNotExist

        with self.assertRaises(HTTPException) as exc:
            await get_verification(VerificationId("v:1"), self.verification_client)

        self.assertEqual(exc.exception.status_code, 404)
        self.assertEqual(exc.exception.detail, "Verification not found")
        self.assertEqual(self.verification_client.mock_calls, [call.get(VerificationId("v:1"))])
