import json
from typing import Any
from unittest.mock import call

from regal import Address
from verifier import ContactRepresentative, Verification
from verifier.exceptions import IncorrectStatus, MethodNotApplicable, VerificationDoesNotExist

from inquisitor.constants import (
    LogEntryType,
    LoggerResult,
    LogReferenceType,
    MessageType,
    VerificationMethod,
)
from inquisitor.main import app
from inquisitor.session import Session
from inquisitor.settings import Settings, get_settings
from inquisitor.token import get_verification

from .utils import (
    CONTACT_ID,
    VERIFICATION_ID,
    ApiTestCase,
    make_event,
    make_verification,
    make_verification_result,
)


class GetTest(ApiTestCase):
    async def test_get(self):
        verification = make_verification()
        self.verification_client.get.return_value = verification
        response = self.client.get("/verification/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Verification(**response.json()), verification)
        self.assertEqual(self.verification_client.mock_calls, [call.get(VERIFICATION_ID)])


class GetMethodsTest(ApiTestCase):
    async def test_get_methods(self):
        methods = ["eID", "letter"]
        self.verification_client.get_methods.return_value = methods
        response = self.client.get("/verification/get_methods/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), methods)
        self.assertEqual(self.verification_client.mock_calls, [call.get_methods(VERIFICATION_ID)])

    async def test_get_methods_verification_not_found(self):
        self.verification_client.get_methods.side_effect = VerificationDoesNotExist
        response = self.client.get("/verification/get_methods/")
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(self.verification_client.mock_calls, [call.get_methods(VERIFICATION_ID)])


class UseMethodTest(ApiTestCase):
    async def _test_use_method(self, status_code: int, response_json: Any) -> None:
        response = self.client.post("/verification/use_method/letter/")
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), response_json)
        self.assertEqual(
            self.verification_client.mock_calls, [call.use_method(VERIFICATION_ID, "letter")]
        )

    async def test_use_method(self):
        verification = make_verification()
        self.verification_client.use_method.return_value = verification
        await self._test_use_method(200, json.loads(verification.json()))

    async def test_use_method_incorrect_status(self):
        self.verification_client.use_method.side_effect = IncorrectStatus
        await self._test_use_method(400, {"detail": "Incorrect verification status"})

    async def test_use_method_method_not_applicable(self):
        self.verification_client.use_method.side_effect = MethodNotApplicable
        await self._test_use_method(400, {"detail": "Method not applicable"})

    async def test_use_method_verification_not_found(self):
        self.verification_client.use_method.side_effect = VerificationDoesNotExist
        await self._test_use_method(404, {"detail": "Verification not found"})


class VerifyTokenTest(ApiTestCase):
    async def _test_verify_token(self, status_code: int, response_json: Any) -> None:
        response = self.client.post("/verification/verify_token/", json={"token": "Gazpacho!"})
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), response_json)
        self.assertEqual(
            self.verification_client.mock_calls,
            [call.verify_token(VERIFICATION_ID, "Gazpacho!", check=False)],
        )

    async def test_verify_token(self):
        self.verification_client.verify_token.return_value = True
        await self._test_verify_token(200, True)

    async def test_verify_token_incorrect_status(self):
        self.verification_client.verify_token.side_effect = IncorrectStatus
        await self._test_verify_token(400, {"detail": "Incorrect verification status"})

    async def test_verify_token_verification_not_found(self):
        self.verification_client.verify_token.side_effect = VerificationDoesNotExist
        await self._test_verify_token(404, {"detail": "Verification not found"})


class VerifyTokenByRepresentativeTest(ApiTestCase):
    async def _test_verify_token_by_representative(
        self, status_code: int, response_json: Any
    ) -> None:
        self.session = Session(verification_id=VERIFICATION_ID, representative_token="Gazpacho!")
        response = self.client.post("/verification/verify_token/representative/")
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), response_json)
        self.assertEqual(
            self.verification_client.mock_calls,
            [call.verify_token(VERIFICATION_ID, "Gazpacho!", check=False)],
        )

    async def test_verify_token_by_representative(self):
        self.verification_client.verify_token.return_value = True
        await self._test_verify_token_by_representative(200, True)

    async def test_verify_token_by_representative_incorrect_status(self):
        self.verification_client.verify_token.side_effect = IncorrectStatus
        await self._test_verify_token_by_representative(
            400, {"detail": "Incorrect verification status"}
        )

    async def test_verify_token_by_representative_verification_not_found(self):
        self.verification_client.verify_token.side_effect = VerificationDoesNotExist
        await self._test_verify_token_by_representative(404, {"detail": "Verification not found"})


class ListVerificationResultsTest(ApiTestCase):
    async def test_list_verification_results(self):
        result = make_verification_result()

        self.verification_client.list_verification_results.return_value = [result]
        response = self.client.get("/verification/results/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [json.loads(result.json())])
        self.assertEqual(
            self.verification_client.mock_calls, [call.list_verification_results(VERIFICATION_ID)]
        )

    async def test_list_verification_results_verification_not_found(self):
        self.verification_client.list_verification_results.side_effect = VerificationDoesNotExist
        response = self.client.get("/verification/results/")
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(
            self.verification_client.mock_calls, [call.list_verification_results(VERIFICATION_ID)]
        )


class RequestManualCheckTest(ApiTestCase):
    async def _test_request_manual_check(self, status_code: int, response_json: Any) -> None:
        response = self.client.post("/verification/request_manual_check/")
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), response_json)
        self.assertEqual(
            self.verification_client.mock_calls, [call.request_manual_check(VERIFICATION_ID)]
        )

    async def test_request_manual_check(self):
        self.verification_client.request_manual_check.return_value = None
        await self._test_request_manual_check(200, None)

    async def test_request_manual_check_incorrect_status(self):
        self.verification_client.request_manual_check.side_effect = IncorrectStatus
        await self._test_request_manual_check(400, {"detail": "Incorrect verification status"})

    async def test_request_manual_check_verification_not_found(self):
        self.verification_client.request_manual_check.side_effect = VerificationDoesNotExist
        await self._test_request_manual_check(404, {"detail": "Verification not found"})


class ResendLetter:
    class BaseTest(ApiTestCase):
        url: str
        message_type: MessageType
        verification_method: VerificationMethod
        method_error_message: str

        async def _test_resend_letter(
            self,
            status_code: int,
            response_json: Any,
            log_result: LoggerResult,
            output_properties: dict,
            mock_calls: list,
        ) -> None:
            response = self.client.post(self.url)
            self.assertEqual(response.json(), response_json)
            self.assertEqual(response.status_code, status_code)
            self.assertLog(
                LogEntryType.RESEND_MESSAGE,
                log_result,
                input_properties={"type": self.message_type.value},
                input_references={
                    LogReferenceType.VERIFICATION: VERIFICATION_ID,
                    LogReferenceType.CONTACT: CONTACT_ID,
                },
                output_properties=output_properties,
            )
            self.assertEqual(self.verification_client.mock_calls, mock_calls)

        async def test_resend_ok(self):
            verification = make_verification(
                events=[make_event(type="use_method", method=self.verification_method)]
            )
            app.dependency_overrides[get_verification] = lambda: verification
            self.verification_client.resend_message.return_value = None
            await self._test_resend_letter(
                status_code=200,
                response_json=None,
                log_result=LoggerResult.SUCCESS,
                output_properties={"message_sent": str(True)},
                mock_calls=[call.resend_message(VERIFICATION_ID, message_type=self.message_type)],
            )

        async def test_resend_method_not_used(self):
            verification = make_verification(events=[])
            app.dependency_overrides[get_verification] = lambda: verification
            await self._test_resend_letter(
                status_code=400,
                response_json={"detail": self.method_error_message},
                log_result=LoggerResult.FAIL,
                output_properties={
                    "message_sent": str(False),
                    "error": self.method_error_message,
                },
                mock_calls=[],
            )

        async def test_resend_limit_exceeded(self):
            verification = make_verification(
                events=[
                    make_event(type="use_method", method=self.verification_method),
                    make_event(type="send_message", method=self.verification_method),
                    make_event(type="send_message", method=self.verification_method),
                ],
            )
            app.dependency_overrides[get_verification] = lambda: verification
            app.dependency_overrides[get_settings] = lambda: Settings(max_letters=2)
            await self._test_resend_letter(
                status_code=400,
                response_json={"detail": "Letter limit exceeded"},
                log_result=LoggerResult.FAIL,
                output_properties={
                    "message_sent": str(False),
                    "error": "Letter limit exceeded",
                },
                mock_calls=[],
            )

        async def test_resend_incorrect_status(self):
            verification = make_verification(
                events=[make_event(type="use_method", method=self.verification_method)]
            )
            app.dependency_overrides[get_verification] = lambda: verification
            self.verification_client.resend_message.side_effect = IncorrectStatus
            await self._test_resend_letter(
                status_code=400,
                response_json={"detail": "Incorrect verification status"},
                log_result=LoggerResult.FAIL,
                output_properties={
                    "message_sent": str(False),
                    "error": "Incorrect verification status",
                },
                mock_calls=[call.resend_message(VERIFICATION_ID, message_type=self.message_type)],
            )

        async def test_resend_verification_not_found(self):
            verification = make_verification(
                events=[make_event(type="use_method", method=self.verification_method)]
            )
            app.dependency_overrides[get_verification] = lambda: verification
            self.verification_client.resend_message.side_effect = VerificationDoesNotExist
            await self._test_resend_letter(
                status_code=404,
                response_json={"detail": "Verification not found"},
                log_result=LoggerResult.FAIL,
                output_properties={
                    "message_sent": str(False),
                    "error": "Verification not found",
                },
                mock_calls=[call.resend_message(VERIFICATION_ID, message_type=self.message_type)],
            )


class ResendConfirmationLetterTest(ResendLetter.BaseTest):
    url = "/verification/letter/resend/"
    message_type = MessageType.CONFIRMATION_LETTER
    verification_method = VerificationMethod.LETTER
    method_error_message = "Letter method is not used"


class ResendRepresentativeLetterTest(ResendLetter.BaseTest):
    url = "/verification/representative/resend/"
    message_type = MessageType.REPRESENTATIVE_LETTER
    verification_method = VerificationMethod.REPRESENTATIVE
    method_error_message = "Representative method is not used"


class SetContactRepresentativeTest(ApiTestCase):
    async def _test_set_contact_representative(
        self,
        status_code: int,
        response_json: Any,
        *,
        log_result: LoggerResult = LoggerResult.SUCCESS,
        output_properties: dict
    ) -> None:
        address = Address(
            street=["Deck 5"],
            city="Red Dwarf",
            postal_code="12345",
            country_code="JMC",
        )
        representative = ContactRepresentative(
            name="Arnold Rimmer",
            organization="Jupiter Mining Corporation",
            address=address,
            telephone="123456",
            email="rimmer@example.com",
        )
        response = self.client.post("/verification/representative/", json=representative.dict())
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.json(), response_json)
        self.assertEqual(
            self.verification_client.mock_calls,
            [call.set_contact_representative(VERIFICATION_ID, representative=representative)],
        )
        self.assertLog(
            LogEntryType.SET_REPRESENTATIVE,
            log_result,
            input_properties={
                "name": "Arnold Rimmer",
                "organization": "Jupiter Mining Corporation",
                "address": str(address),
                "telephone": "123456",
                "email": "rimmer@example.com",
            },
            input_references={
                LogReferenceType.VERIFICATION: VERIFICATION_ID,
                LogReferenceType.CONTACT: CONTACT_ID,
            },
            output_properties=output_properties,
        )

    async def test_set_contact_representative(self):
        self.verification_client.set_contact_representative.return_value = None
        await self._test_set_contact_representative(200, None, output_properties={})

    async def test_set_contact_representative_incorrect_status(self):
        self.verification_client.set_contact_representative.side_effect = IncorrectStatus
        await self._test_set_contact_representative(
            400,
            {"detail": "Incorrect verification status"},
            log_result=LoggerResult.FAIL,
            output_properties={"error": "Incorrect verification status"},
        )

    async def test_set_contact_representative_verification_not_found(self):
        self.verification_client.set_contact_representative.side_effect = VerificationDoesNotExist
        await self._test_set_contact_representative(
            404,
            {"detail": "Verification not found"},
            log_result=LoggerResult.FAIL,
            output_properties={"error": "Verification not found"},
        )
