"""Token type and related functions."""
from datetime import datetime, timedelta
from typing import cast

from fastapi import Depends, HTTPException, Request, Security, status
from fastapi.openapi.models import OAuthFlows as OAuthFlowsModel
from fastapi.security import OAuth2, SecurityScopes
from fastapi.security.utils import get_authorization_scheme_param
from fred_types import ContactId
from jose import JWTError, jwe, jwt
from jose.constants import ALGORITHMS
from jose.exceptions import JWEError
from pydantic import BaseModel
from verifier import Verification, VerificationClient, VerificationId
from verifier.exceptions import VerificationDoesNotExist

from .constants import OAUTH_COOKIE, Scope
from .settings import Settings, get_settings, get_verification_client


class OAuth2PasswordBearerWithCookie(OAuth2):
    """OAuth bearer stored in cookie."""

    def __init__(self, *, tokenUrl: str, scopes: dict, scheme_name: str | None = None):
        flows = OAuthFlowsModel(password={"tokenUrl": tokenUrl, "scopes": scopes})
        super().__init__(flows=flows, scheme_name=scheme_name)

    async def __call__(self, request: Request) -> str:
        """Get access token from cookie."""
        authorization: str | None = request.cookies.get(OAUTH_COOKIE)

        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Not authenticated",
                headers={"WWW-Authenticate": "Bearer"},
            )

        return param


oauth2_scheme = OAuth2PasswordBearerWithCookie(
    tokenUrl="auth/token/",
    scopes={
        Scope.READ: "Read verification data",
        Scope.WRITE: "Perform operations with verification.",
    },
)


class TokenData(BaseModel):
    """Auth token data."""

    verification_id: VerificationId
    contact_id: ContactId
    scopes: list[Scope] = []


def create_access_token(data: TokenData, settings: Settings) -> str:
    """Create access JWT."""
    to_encode = data.dict()
    to_encode.update(
        {
            # issued at
            "iat": datetime.utcnow(),
            # not valid before
            "nbf": datetime.utcnow() - timedelta(minutes=5),
            # expiration time
            "exp": datetime.utcnow() + timedelta(seconds=settings.access_token_expires),
        }
    )
    encoded_jwt = jwt.encode(to_encode, settings.secret_key, algorithm=ALGORITHMS.HS256)
    encrypted_jwt = jwe.encrypt(
        encoded_jwt,
        key=settings.secret_key_sha256,
        algorithm=ALGORITHMS.DIR,
        encryption=ALGORITHMS.A256GCM,
    ).decode()
    return cast(str, encrypted_jwt)


async def auth_token_data(
    security_scopes: SecurityScopes,
    token: str = Depends(oauth2_scheme),
    settings: Settings = Depends(get_settings),
) -> TokenData:
    """Get token data from encrypted JWT."""
    try:
        decrypted_jwt = jwe.decrypt(token, settings.secret_key_sha256)
        payload = jwt.decode(decrypted_jwt, settings.secret_key, algorithms=[ALGORITHMS.HS256])
        data = TokenData(**payload)

        # Check security scopes
        if set(security_scopes.scopes) <= set(data.scopes):
            return data
        else:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="Insufficient permissions",
                headers={"WWW-Authenticate": f'Bearer scope="{security_scopes.scope_str}"'},
            )
    except (JWTError, JWEError):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )


async def get_contact_id(
    data: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> ContactId:
    """Get contact id."""
    return data.contact_id


async def get_verification_id(
    data: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> VerificationId:
    """Get verification id."""
    return data.verification_id


async def get_verification(
    verification_id: VerificationId = Security(get_verification_id, scopes=[Scope.READ]),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> Verification:
    """Get verification."""
    try:
        return await verification_client.get(verification_id)
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")
