"""Verification endpoints."""
from typing import Any

from fastapi import APIRouter, Body, Depends, HTTPException, Security, status
from fred_types import ContactId
from grill import Logger
from verifier import (
    ContactRepresentative,
    Verification,
    VerificationClient,
    VerificationId,
    VerificationResult,
)
from verifier.exceptions import IncorrectStatus, MethodNotApplicable, VerificationDoesNotExist

from .constants import (
    COMMON_RESPONSES,
    EventType,
    LogEntryType,
    LoggerResult,
    LogReferenceType,
    MessageType,
    Scope,
    VerificationMethod,
)
from .models import Error
from .session import Session, get_session
from .settings import Settings, get_logger, get_settings, get_verification_client
from .token import get_contact_id, get_verification, get_verification_id
from .utils import get_source_ip

router = APIRouter(prefix="/verification")

RESPONSES: dict[int | str, dict[str, Any]] = {
    **COMMON_RESPONSES,
    404: {"description": "Verification not found", "model": Error},
}


@router.get("/", responses=RESPONSES)
async def get(verification: Verification = Depends(get_verification)) -> Verification:
    """Get verification info."""
    return verification


@router.get("/get_methods/", responses=RESPONSES)
async def get_methods(
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> list[str]:
    """Get list of applicable verification methods."""
    try:
        return await verification_client.get_methods(verification_id)
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


@router.post(
    "/use_method/{method}/",
    responses={
        **RESPONSES,
        400: {"description": "Bad request", "model": Error},
    },
)
async def use_method(
    method: VerificationMethod,
    verification_id: VerificationId = Security(get_verification_id, scopes=[Scope.WRITE]),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> Verification:
    """Use selected verification method."""
    try:
        return await verification_client.use_method(verification_id, method)
    except IncorrectStatus:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect verification status"
        )
    except MethodNotApplicable:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Method not applicable"
        )
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


@router.post(
    "/verify_token/",
    responses={
        **RESPONSES,
        400: {"description": "Incorrect verification status", "model": Error},
    },
)
async def verify_token(
    token: str = Body(embed=True),
    verification_id: VerificationId = Security(get_verification_id, scopes=[Scope.WRITE]),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> bool:
    """Verify token."""
    try:
        return await verification_client.verify_token(verification_id, token, check=False)
    except IncorrectStatus:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Incorrect verification status",
        )
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


@router.post(
    "/verify_token/representative/",
    responses={
        **RESPONSES,
        400: {"description": "Incorrect verification status", "model": Error},
    },
)
async def verify_token_by_representative(
    verification_id: VerificationId = Security(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
    session: Session = Security(get_session),
) -> bool:
    """Verify token by contact representative."""
    try:
        # Verification token is taken from session data where it was stored during contact
        # representative login.
        token = session.representative_token or ""
        return await verification_client.verify_token(verification_id, token, check=False)
    except IncorrectStatus:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Incorrect verification status",
        )
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


@router.get("/results/", responses=RESPONSES)
async def list_verification_results(
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> list[VerificationResult]:
    """List verification results."""
    try:
        return await verification_client.list_verification_results(verification_id)
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


@router.post(
    "/request_manual_check/",
    responses={
        **RESPONSES,
        400: {"description": "Incorrect verification status", "model": Error},
    },
)
async def request_manual_check(
    verification_id: VerificationId = Security(get_verification_id, scopes=[Scope.WRITE]),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> None:
    """Request manual check of verification data."""
    try:
        await verification_client.request_manual_check(verification_id)
    except IncorrectStatus:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect verification status"
        )
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


async def _resend_letter(
    verification: Verification,
    source_ip: str | None,
    logger: Logger,
    settings: Settings,
    verification_client: VerificationClient,
    message_type: MessageType,
    verification_method: VerificationMethod,
    method_error_message: str,
) -> None:
    log_entry = logger.create(
        LogEntryType.RESEND_MESSAGE,
        source_ip=source_ip,
        properties={
            "type": message_type.value,
        },
        references={
            LogReferenceType.VERIFICATION: verification.verification_id,
            LogReferenceType.CONTACT: verification.contact_id,
        },
    )
    with log_entry:
        log_entry.properties["message_sent"] = str(False)

        # Check whether right verification method has been used
        is_method_used = any(
            event.type == EventType.USE_METHOD and event.method == verification_method
            for event in verification.events
        )
        if not is_method_used:
            log_entry.properties["error"] = method_error_message
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail=method_error_message
            )

        # Check whether max number of letters has been exceeded
        letters = [
            event
            for event in verification.events
            if event.type == EventType.SEND_MESSAGE and event.method == verification_method
        ]
        if len(letters) >= settings.max_letters:
            log_entry.properties["error"] = "Letter limit exceeded"
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Letter limit exceeded"
            )

        try:
            await verification_client.resend_message(
                verification.verification_id,
                message_type=message_type,
            )
            log_entry.properties["message_sent"] = str(True)
            log_entry.result = LoggerResult.SUCCESS
        except VerificationDoesNotExist:
            log_entry.properties["error"] = "Verification not found"
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found"
            )
        except IncorrectStatus:
            log_entry.properties["error"] = "Incorrect verification status"
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect verification status"
            )


@router.post(
    "/letter/resend/",
    responses={
        **RESPONSES,
        400: {"description": "Incorrect status", "model": Error},
    },
)
async def resend_confirmation_letter(
    verification: Verification = Security(get_verification, scopes=[Scope.WRITE]),
    source_ip: str | None = Depends(get_source_ip),
    logger: Logger = Depends(get_logger),
    settings: Settings = Depends(get_settings),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> None:
    """Resend confirmation letter.

    There is a limit of maximum number of letters sent per verification.
    """
    await _resend_letter(
        verification=verification,
        source_ip=source_ip,
        logger=logger,
        settings=settings,
        verification_client=verification_client,
        message_type=MessageType.CONFIRMATION_LETTER,
        verification_method=VerificationMethod.LETTER,
        method_error_message="Letter method is not used",
    )


@router.post(
    "/representative/resend/",
    responses={
        **RESPONSES,
        400: {"description": "Incorrect status", "model": Error},
    },
)
async def resend_representative_letter(
    verification: Verification = Security(get_verification, scopes=[Scope.WRITE]),
    source_ip: str | None = Depends(get_source_ip),
    logger: Logger = Depends(get_logger),
    settings: Settings = Depends(get_settings),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> None:
    """Resend confirmation letter to contact representative.

    There is a limit of maximum number of letters sent per verification.
    """
    await _resend_letter(
        verification=verification,
        source_ip=source_ip,
        logger=logger,
        settings=settings,
        verification_client=verification_client,
        message_type=MessageType.REPRESENTATIVE_LETTER,
        verification_method=VerificationMethod.REPRESENTATIVE,
        method_error_message="Representative method is not used",
    )


@router.post(
    "/representative/",
    responses={
        **RESPONSES,
        400: {"description": "Bad request", "model": Error},
    },
)
async def set_contact_representative(
    representative: ContactRepresentative,
    source_ip: str | None = Depends(get_source_ip),
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
    contact_id: ContactId = Depends(get_contact_id),
    logger: Logger = Depends(get_logger),
) -> None:
    """Set contact representative.

    If you wish to start verification using this representative,
    you must call ``use_method`` endpoint afterwards.
    """
    log_entry = logger.create(
        LogEntryType.SET_REPRESENTATIVE,
        source_ip=source_ip,
        properties={k: str(v) for k, v in representative},
        references={
            LogReferenceType.VERIFICATION: verification_id,
            LogReferenceType.CONTACT: contact_id,
        },
    )
    with log_entry:
        try:
            await verification_client.set_contact_representative(
                verification_id,
                representative=representative,
            )
        except VerificationDoesNotExist:
            log_entry.properties["error"] = "Verification not found"
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found"
            )
        except IncorrectStatus:
            log_entry.properties["error"] = "Incorrect verification status"
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect verification status"
            )

        log_entry.result = LoggerResult.SUCCESS
